Type TAccountSystem
	Const ACCOUNT_FILE:String = "Accounts.dat"
	
	Global RegisteredPlayers:TRegisteredPlayer[]
	Global NameToPassword:TMap
	
	Function Init()
		NameToPassword = CreateMap()
		
		LoadAccounts()
	End Function
	
	Function LoadAccounts()
		Local Instream:TStream = ReadFile( ACCOUNT_FILE )
		
		If Instream Then
			While Not Instream.Eof()
				Local PlayerName:String = Instream.ReadLine()
				Local PlayerPass:String = Instream.ReadLine()
				
				RegisteredPlayers :+ [ TRegisteredPlayer.Create( PlayerName, PlayerPass ) ]
				
				NameToPassword.Insert( PlayerName, PlayerPass )
			Wend
			
			Instream.Close()
		EndIf
	End Function
	
	Function SaveAccounts()
		Local Outstream:TStream = WriteFile( ACCOUNT_FILE )
		
		If Outstream Then
			For Local RegisteredPlayer:TRegisteredPlayer = EachIn RegisteredPlayers
				Outstream.WriteLine( RegisteredPlayer.Name )
				Outstream.WriteLine( RegisteredPlayer.Pass )
			Next
			
			Outstream.Close()
		EndIf
	End Function
	
	Function AddPlayer( Name:String, Pass:String )
		RegisteredPlayers :+ [ TRegisteredPlayer.Create( Name, Pass ) ]
		
		NameToPassword.Insert( Name, Pass )
		
		SaveAccounts()
	End Function
	
	Function ChangePlayerPassword( Name:String, Pass:String )
		For Local Player:TRegisteredPlayer = EachIn RegisteredPlayers
			If Player.Name.ToLower() = Name Then
				Player.Pass = Pass
				
				Exit
			EndIf
		Next
		
		NameToPassword.Insert( Name, Pass )
		
		SaveAccounts()
	End Function
	
	Function RemovePlayer( Name:String )
		NameToPassword.Remove( Name )
		
		For Local I:Int = 0 Until RegisteredPlayers.Length
			If RegisteredPlayers[ I ].Name.ToLower() = Name Then
				RegisteredPlayers = RegisteredPlayers[ .. I ] + RegisteredPlayers[ I + 1 .. ]
				
				SaveAccounts()
				
				Return
			EndIf
		Next
	End Function
	
	Function PlayerExists:Int( Name:String )
		Return NameToPassword.ValueForKey( Name ) <> Null
	End Function
	
	Function CorrectPassword:Int( Name:String, Pass:String )
		Return String( NameToPassword.ValueForKey( Name ) ) = Pass
	End Function
End Type

Type TRegisteredPlayer
	Field Name:String
	Field Pass:String
	
	Function Create:TRegisteredPlayer( Name:String, Pass:String )
		Local RegisteredPlayer:TRegisteredPlayer = New TRegisteredPlayer
		
		RegisteredPlayer.Name = Name
		RegisteredPlayer.Pass = Pass
		
		Return RegisteredPlayer
	End Function
End Type