Import BRL.FileSystem

Type TConfiguration
	Field InBandwidth:Int
	Field OutBandwidth:Int
	Field MaxPlayers:Int
	Field MapWidth:Int
	Field MapHeight:Int
	Field MapDepth:Int
	Field Port:Int
	Field AdminPassword:String
	Field LoadMap:String
	Field AutoSaveDelay:Int
	
	Function Create:TConfiguration( InBandwidth:Int, OutBandwidth:Int, MaxPlayers:Int, MapWidth:Int, MapHeight:Int, MapDepth:Int, Port:Int, AdminPassword:String, LoadMap:String, AutoSaveDelay:Int )
		Local Config:TConfiguration = New TConfiguration
		
		Config.InBandwidth = InBandwidth
		Config.OutBandwidth = OutBandwidth
		Config.MaxPlayers = MaxPlayers
		Config.MapWidth = MapWidth
		Config.MapHeight = MapHeight
		Config.MapDepth = MapDepth
		Config.Port = Port
		Config.AdminPassword = AdminPassword
		Config.LoadMap = LoadMap
		Config.AutoSaveDelay = AutoSaveDelay
		
		Return Config
	End Function
End Type

Type TConfigurationService
	Const FILENAME:String = "Config.ini"
	
	Global Config:TConfiguration = TConfiguration.Create( 0, 0, 64, 512, 64, 512, 16000, MilliSecs(), "", 900 )
	
	Function LoadConfigFile()
		Local Instream:TStream = ReadFile( FILENAME )
		
		If Instream Then
			While Not Instream.Eof()
				Local Line:String = Instream.ReadLine()
				
				Local Tokens:String[] = Line.Split( "=" )
				
				Select Tokens[ 0 ].Trim().ToLower()
					Case "inbandwidth"
						Config.InBandwidth = Tokens[ 1 ].Trim().ToInt()
					Case "outbandwidth"
						Config.OutBandwidth = Tokens[ 1 ].Trim().ToInt()
					Case "maxplayers"
						Config.MaxPlayers = Tokens[ 1 ].Trim().ToInt()
					Case "mapwidth"
						Config.MapWidth = Tokens[ 1 ].Trim().ToInt()
					Case "mapheight"
						Config.MapHeight = Tokens[ 1 ].Trim().ToInt()
					Case "mapdepth"
						Config.MapDepth = Tokens[ 1 ].Trim().ToInt()
					Case "port"
						Config.Port = Tokens[ 1 ].Trim().ToInt()
					Case "adminpassword"
						Config.AdminPassword = Tokens[ 1 ].Trim()
					Case "loadmap"
						Config.LoadMap = Tokens[ 1 ].Trim()
					Case "autosavedelay"
						Config.AutoSaveDelay = Tokens[ 1 ].Trim().ToInt()
				End Select
			Wend
			
			Instream.Close()
		Else
			SaveConfigFile()
		EndIf
	End Function
	
	Function SaveConfigFile()
		Local Outstream:TStream = WriteFile( FILENAME )
		
		Outstream.WriteLine( "Port = " + Config.Port )
		Outstream.WriteLine( "InBandwidth = " + Config.InBandwidth )
		Outstream.WriteLine( "OutBandwidth = " + Config.OutBandwidth )
		Outstream.WriteLine( "MaxPlayers = " + Config.MaxPlayers )
		Outstream.WriteLine( "MapWidth = " + Config.MapWidth )
		Outstream.WriteLine( "MapHeight = " + Config.MapHeight )
		Outstream.WriteLine( "MapDepth = " + Config.MapDepth )
		Outstream.WriteLine( "AdminPassword = " + Config.AdminPassword )
		Outstream.WriteLine( "LoadMap = " + Config.LoadMap )
		Outstream.WriteLine( "AutoSaveDelay = " + Config.AutoSaveDelay )
		
		Outstream.Close()
	End Function
	
	Function GetConfiguration:TConfiguration()
		Return Config
	End Function
	
	Function SetConfiguration( Configuration:TConfiguration )
		Config = Configuration
	End Function
End Type
