SuperStrict

Import BRL.Stream
Import BRL.SocketStream
Import BRL.Bank
Import BRL.StandardIO

Import "enet-1.2.2/include/*.h"
Import "enet-1.2.2/callbacks.c"
Import "enet-1.2.2/host.c"
Import "enet-1.2.2/list.c"
Import "enet-1.2.2/packet.c"
Import "enet-1.2.2/peer.c"
Import "enet-1.2.2/protocol.c"
?Win32
	Import "enet-1.2.2/win32.c"
?Not Win32
	Import "enet-1.2.2/unix.c"
?

Extern
	Function ENet_Initialize() = "enet_initialize"
	Function ENet_DeInitialize() = "enet_deinitialize"
	
	Function ENet_HostCreate:Byte Ptr( Address:Byte Ptr, PeerCount:Int, BandwidthIn:Int, BandwidthOut:Int ) = "enet_host_create"
	Function ENet_HostService:Int( Host:Byte Ptr, Event:Byte Ptr, Timeout:Int ) = "enet_host_service"
	Function ENet_HostDestroy( Host:Byte Ptr ) = "enet_host_destroy"
	Function ENet_HostFlush( Host:Byte Ptr ) = "enet_host_flush"
	Function ENet_HostConnect:Byte Ptr( Host:Byte Ptr, Address:Byte Ptr, ChannelCount:Int ) = "enet_host_connect"
	
	Function ENet_PacketDestroy( Packet:Byte Ptr ) = "enet_packet_destroy"
	Function ENet_PacketCreate:Byte Ptr( Data:Byte Ptr, Size:Int, Flags:Int ) = "enet_packet_create"
	Function ENet_PacketResize:Int( Packet:Byte Ptr, Size:Int ) = "enet_packet_resize"
	
	Function ENet_PeerSend:Int( Peer:Byte Ptr, ChannelID:Byte, Packet:Byte Ptr ) = "enet_peer_send"
	Function ENet_PeerDisconnect( Peer:Byte Ptr, Data:Int ) = "enet_peer_disconnect"
	Function ENet_PeerReset( Peer:Byte Ptr ) = "enet_peer_reset"
	
	Function ENet_AddressSetHost( Address:Byte Ptr, HostName:Byte Ptr ) = "enet_address_set_host"
	Function ENet_AddressGetHost( Address:Byte Ptr, HostName:Byte Ptr, NameLength:Int ) = "enet_address_get_host"
End Extern

Type TENetHost
	Field Original:Byte Ptr
	
	Field LastEvent:TENetEvent
		
	Method Delete()
		If Original <> Null Then ENet_HostDestroy( Original )
	End Method
	
	Method GetLastEvent:TENetEvent()
		Return LastEvent
	End Method
	
	Method Service:Int( Timeout:Int )
		Local Event:TENetEvent = New TENetEvent
		
		Local Result:Int = ENet_HostService( Original, Event.ToENetEvent(), Timeout )
		
		If Result > 0 Then
			Event.FromENetEvent()
			
			LastEvent = Event
		EndIf
		
		Return Result
	End Method
	
	Method Connect:TENetPeer( Address:TENetAddress, ChannelCount:Int )
		Local PeerPtr:Byte Ptr = ENet_HostConnect( Original, Address.ToENetAddress(), ChannelCount )
		
		If PeerPtr = Null Then Return Null
		
		Return TENetPeer.CreateFromENetPeer( PeerPtr )
	End Method
	
	Method Flush()
		ENet_HostFlush( Original )
	End Method
	
	Function Create:TENetHost( Address:TENetAddress, PeerCount:Int, BandwidthIn:Int, BandwidthOut:Int )
		Local Host:TENetHost = New TENetHost
		
		If Address Then
			Host.Original = ENet_HostCreate( Address.ToENetAddress(), PeerCount, BandwidthIn, BandwidthOut )
		Else
			Host.Original = ENet_HostCreate( Null, PeerCount, BandwidthIn, BandwidthOut )
		EndIf
		
		If Host.Original = Null Then
			Return Null
		Else
			Return Host
		EndIf
	End Function
End Type

Type TENetPeer
	Const ENET_PEER_PACKET_LOSS_SCALE:Int = ( 1 Shl 16 )
	
	Field Original:Byte Ptr
	
	Field Address:TENetAddress
	Field InBandwidth:Int
	Field OutBandwidth:Int
	Field PacketLoss:Float
	Field Ping:Int
	
	Method ToENetPeer:Byte Ptr()
		Return Original
	End Method
	
	Method FromENetPeer()
		Address = TENetAddress.CreateFromENetAddress( Original + 20 )
		InBandwidth  = ( Int Ptr ( Original + 44  ) )[ 0 ]
		OutBandwidth = ( Int Ptr ( Original + 48  ) )[ 0 ]
		PacketLoss   = ( Int Ptr ( Original + 96  ) )[ 0 ]/Float( ENET_PEER_PACKET_LOSS_SCALE )
		Ping         = ( Int Ptr ( Original + 148 ) )[ 0 ]
	End Method
	
	Method Send:Int( Packet:TENetPacket, ChannelID:Byte = 0 )
		Local Result:Int = ENet_PeerSend( Original, ChannelID, Packet.ToENetPacket() )
		
		Packet.FlagNotDelete()
		
		Return Result
	End Method
	
	Method Disconnect( Data:Int = 0 )
		ENet_PeerDisconnect( Original, Data )
	End Method
	
	Method Reset()
		ENet_PeerReset( Original )
	End Method
	
	Function CreateFromENetPeer:TENetPeer( Original:Byte Ptr )
		If Original = Null Then Return Null
		
		Local Peer:TENetPeer = New TENetPeer
		
		Peer.Original = Original
		
		Peer.FromENetPeer()
		
		Return Peer
	End Function
End Type

Type TENetAddress
	Const ENET_HOST_ANY      :Int =  0
	Const ENET_HOST_BROADCAST:Int = $FFFFFFFF
	Const ENET_PORT_ANY      :Int =  0
	
	Field Original:Short[]
	
	Field Host:Int
	Field Port:Short
	
	Method New()
		Original = New Short[ 3 ]
	End Method
	
	Method FromENetAddress()
		Host = ( Original[ 1 ] Shl 16 ) | Original[ 0 ]
		Port = Original[ 2 ]
	End Method
	
	Method ToENetAddress:Byte Ptr()
		Original[ 0 ] = Host & $FFFF
		Original[ 1 ] = ( Host Shr 16 ) & $FFFF
		Original[ 2 ] = Port
		
		Return Varptr Original[ 0 ]
	End Method
	
	Method ToString:String()
		Local Buffer:Byte Ptr = MemAlloc( 64 )
		
		ENet_AddressGetHost( ToENetAddress(), Buffer, 64 )
		
		Local Result:String = String.FromCString( Buffer )
		
		MemFree Buffer
		
		Return Result
	End Method
	
	Method SetHostFromString( HostName:String )
		Local Buffer:Byte Ptr = HostName.ToCString()
		
		ENet_AddressSetHost( Varptr Original[ 0 ], Buffer )
		
		MemFree( Buffer )
		
		Host = ( Original[ 1 ] Shl 2 ) + Original[ 0 ]
	End Method
	
	Function CreateFromENetAddress:TENetAddress( Original:Byte Ptr )
		Local Address:TENetAddress = New TENetAddress
		
		Address.Original[ 0 ] = ( Short Ptr Original )[ 0 ]
		Address.Original[ 1 ] = ( Short Ptr Original )[ 1 ]
		Address.Original[ 2 ] = ( Short Ptr Original )[ 2 ]
		Address.FromENetAddress()
		
		Return Address
	End Function
	
	Function Create:TENetAddress( Host:Int, Port:Short )
		Local Address:TEnetAddress = New TEnetAddress
		
		Address.Host = Host
		Address.Port = Port
		
		Return Address
	End Function
End Type

Type TENetEvent
	Const ENET_EVENT_TYPE_NONE      :Int = 0
	Const ENET_EVENT_TYPE_CONNECT   :Int = 1
	Const ENET_EVENT_TYPE_DISCONNECT:Int = 2
	Const ENET_EVENT_TYPE_RECEIVE   :Int = 3
	
	Field OriginalBank:TBank
	Field Original:Byte Ptr

	Field EventType:Int
	Field Peer:TENetPeer
	Field ChannelID:Byte
	Field Data:Int
	Field Packet:TENetPacket
	
	Method New()
		OriginalBank = CreateBank( 17 )
		Original = OriginalBank.Lock()
		
		For Local I:Int = 0 Until 17
			Original[ I ] = 0
		Next
	End Method
	
	Method ToENetEvent:Byte Ptr()
		               (  Int Ptr     ( Original + 0  ) )[ 0 ] = EventType
		If Peer Then   ( Byte Ptr Ptr ( Original + 4  ) )[ 0 ] = Peer.ToENetPeer()
		               ( Byte Ptr     ( Original + 8  ) )[ 0 ] = ChannelID
		               (  Int Ptr     ( Original + 9  ) )[ 0 ] = Data
		If Packet Then ( Byte Ptr Ptr ( Original + 13 ) )[ 0 ] = Packet.ToENetPacket()
		
		Return Original
	End Method
	
	Method FromENetEvent()
		Local PeerPtr:Byte Ptr, PacketPtr:Byte Ptr
		
		EventType = (  Int Ptr     ( Original + 0  ) )[ 0 ]
		PeerPtr   = ( Byte Ptr Ptr ( Original + 4  ) )[ 0 ]
		ChannelID = ( Byte Ptr     ( Original + 8  ) )[ 0 ]
		Data      = (  Int Ptr     ( Original + 9  ) )[ 0 ]
		PacketPtr = ( Byte Ptr Ptr ( Original + 16 ) )[ 0 ]
		
		Peer   = TENetPeer  .CreateFromENetPeer  ( PeerPtr )
		Packet = TENetPacket.CreateFromENetPacket( PacketPtr )
	End Method
End Type

Type TENetStream Extends TStream
	Field Packet:TENetPacket
	Field Position:Int
	
	Method Pos:Int()
		Return Position
	End Method
	
	Method Size:Int()
		Return Packet.Size
	End Method
	
	Method Seek( Pos:Int )
		Position = Pos
	End Method
	
	Method Eof:Int()
		Return Position >= Packet.Size
	End Method
	
	Method Read:Int( Dest:Byte Ptr, Count:Int )
		Return ReadBytes( Dest, Count )
	End Method
	
	Method Write:Int( Source:Byte Ptr, Count:Int )
		Return WriteBytes( Source, Count )
	End Method
	
	Method ReadBytes:Int( Dest:Byte Ptr, Count:Int )
		If Position + Count > Packet.Size Then Return 0
		
		MemCopy( Dest, Packet.Data + Position, Count )
		
		Position :+ Count
		
		Return Count
	End Method
	
	Method WriteBytes:Int( Source:Byte Ptr, Count:Int )
		If Position + Count > Packet.Size Then
			If Packet.Resize( Position + Count ) Then Throw New TStreamWriteException
		EndIf
		
		MemCopy( Packet.Data + Position, Source, Count )
		
		Position :+ Count
		
		Return Count
	End Method

	Method SkipBytes:Int( Count:Int )
		If Position + Count > Packet.Size Then Return 0
		
		Position :+ Count
	End Method
	
	Function Create:TENetStream( Packet:TENetPacket )
		If Not Packet Then Return Null
		
		Local Stream:TENetStream = New TENetStream
		
		Stream.Packet = Packet
		
		Return Stream
	End Function
End Type

Type TENetPacket
	Const ENET_PACKET_FLAG_RELIABLE   :Int = %001
	Const ENET_PACKET_FLAG_UNSEQUENCED:Int = %010
	Const ENET_PACKET_FLAG_NO_ALLOCATE:Int = %100
	
	Field Original:Byte Ptr
	
	Field Size:Int
	Field Data:Byte Ptr
	Field Flags:Int
	
	Field DoDelete:Int
	
	Method Delete()
		If Original <> Null And DoDelete Then ENet_PacketDestroy( Original )
	End Method
	
	Method FlagNotDelete()
		DoDelete = False
	End Method
	
	Method FromENetPacket()
		Flags = ( Int Ptr      ( Original + 4  ) )[ 0 ]
		Data  = ( Byte Ptr Ptr ( Original + 8  ) )[ 0 ]
		Size  = ( Int Ptr      ( Original + 12 ) )[ 0 ]
	End Method
	
	Method ToENetPacket:Byte Ptr()
		Return Original
	End Method
	
	Method Resize:Int( NewSize:Int )
		Local Result:Int = ENet_PacketResize( Original, NewSize )
		
		If Result = 0 Then FromENetPacket()
		
		Return Result
	End Method
	
	Function CreateFromENetPacket:TENetPacket( Original:Byte Ptr )
		If Original = Null Then Return Null
		
		Local Packet:TENetPacket = New TENetPacket
		
		Packet.Original = Original
		
		Packet.FromENetPacket()
		
		Return Packet
	End Function
	
	Function Create:TENetPacket( InitialData:Byte Ptr, PacketSize:Int, Flags:Int )
		Local Packet:TENetPacket = New TENetPacket
		
		Packet.Original = ENet_PacketCreate( InitialData, PacketSize, Flags )
		
		If Packet.Original = Null Then Return Null
		
		Packet.FromENetPacket()
		
		Packet.DoDelete = True
		
		Return Packet
	End Function
End Type