Type TLogger
	Const ACTION_MODIFY:Int = 1
	
	Global PlayerToLoggedPlayer:TMap
	Global Players:TLoggedPlayer[]
	Global PlayerActions:TMap
	
	Function Init()
		PlayerToLoggedPlayer = CreateMap()
		PlayerActions = CreateMap()
	End Function
	
	Function AddPlayer( Player:TPlayer )
		Local LoggedPlayer:TLoggedPlayer = TLoggedPlayer.Create( Player.Name, Player.Peer.Address.ToString() )
		
		Players :+ [ LoggedPlayer ]
		PlayerToLoggedPlayer.Insert( Player, LoggedPlayer )
		PlayerActions.Insert( LoggedPlayer, New TList )
	End Function
	
	Function RemovePlayer( Player:TPlayer )
		PlayerToLoggedPlayer.Remove( Player )
	End Function
	
	Function LogMapModify( Player:TPlayer, X:Int, Y:Int, Z:Int, NewValue:Int )
		Local LoggedPlayer:TLoggedPlayer = TLoggedPlayer( PlayerToLoggedPlayer.ValueForKey( Player ) )
		
		If LoggedPlayer Then
			Local List:TList = TList( PlayerActions.ValueForKey( LoggedPlayer ) )
			
			If List Then List.AddLast( [ ACTION_MODIFY, X, Y, Z, NewValue, time_( Null ) ] )
		EndIf
	End Function
	
	Function Save:Int( Path:String )
		Local Outstream:TStream = WriteFile( Path )
		
		If Outstream Then
			Outstream.WriteInt( Players.Length )
			
			For Local LoggedPlayer:TLoggedPlayer = EachIn Players
				Outstream.WriteLine( LoggedPlayer.Name )
				Outstream.WriteLine( LoggedPlayer.Host )
				
				Local List:TList = TList( PlayerActions.ValueForKey( LoggedPlayer ) )
				
				If List Then
					Outstream.WriteInt( List.Count() )
					
					For Local Action:Int[] = EachIn List
						Select Action[ 0 ]
							Case ACTION_MODIFY
								For Local I:Int = 0 Until 6
									Outstream.WriteInt( Action[ I ] )
								Next
						End Select
					Next
				Else
					Outstream.WriteInt( 0 )
				EndIf
			Next
			
			Outstream.Close()
			
			Return True
		EndIf
	End Function
	
	Function Load:Int( Path:String )
		Local Instream:TStream = ReadFile( Path )
		
		If Instream Then
			Local Count:Int = Instream.ReadInt()
			
			For Local I:Int = 0 Until Count
				Local Name:String = Instream.ReadLine()
				Local Host:String = Instream.ReadLine()
				
				Local LoggedPlayer:TLoggedPlayer = TLoggedPlayer.Create( Name, Host )
				
				Players :+ [ LoggedPlayer ]
				
				Local List:TList = New TList
				
				PlayerActions.Insert( LoggedPlayer, List )
				
				Local ActionCount:Int = Instream.ReadInt()
				
				For Local T:Int = 0 Until ActionCount
					Local ActionType:Int = Instream.ReadInt()
					
					Local Action:Int[] = [ ActionType ]
					
					Select ActionType
						Case ACTION_MODIFY
							For Local S:Int = 1 To 5
								Action :+ [ Instream.ReadInt() ]
							Next
					End Select
					
					List.AddLast( Action )
				Next
			Next
		EndIf
	End Function
	
	Function Clear()
		PlayerToLoggedPlayer.Clear()
		Players = Null
		PlayerActions.Clear()
	End Function
End Type

Type TLoggedPlayer
	Field Name:String
	Field Host:String
	
	Function Create:TLoggedPlayer( Name:String, Host:String )
		Local LoggedPlayer:TLoggedPlayer = New TLoggedPlayer
		
		LoggedPlayer.Name = Name
		LoggedPlayer.Host = Host
		
		Return LoggedPlayer
	End Function
End Type