Type TLuaImporter
	Const BLOCK_PATH:String = "Data/Blocks.lua"
	
	Global L:Byte Ptr
	
	Function Init()
		L = luaL_newstate()
		luaL_openlibs( L )
		
		TBlockType.BlockTypes[ 0 ] = New TBLockType
		
		LoadBlocks()
	End Function
	
	Function LoadBlocks()
		lua_newtable( L )
		lua_setglobal( L, "Blocks" )
		
		If luaL_loadfile( L, BLOCK_PATH ) Then TServer.AbortServer( "Error while loading lua script:" + lua_tostring( L, 1 ) )
		
		If lua_pcall( L, 0, 0, 0 ) Then TServer.AbortServer( "Error while executing lua script:" + lua_tostring( L, 1 ) )
		
		lua_getglobal( L, "Blocks" )
		Local Index:Int = lua_gettop( L )
		
		lua_pushnil( L )
		While lua_next( L, Index )
			LoadBlock()
			
			lua_pop( L, 1 )
		Wend
		lua_pop( L, 1 )
	End Function
	
	Function LoadBlock()
		Local Block:TBlockType = New TBlockType
		
		Local Index:Int = lua_gettop( L )
		
		lua_getfield( L, Index, "ID" )
		lua_getfield( L, Index, "IsObject" )
		lua_getfield( L, Index, "IsTrigger" )
		lua_getfield( L, Index, "TriggeredOnTouch" )
		lua_getfield( L, Index, "TriggeredBlock" )
		lua_getfield( L, Index, "ReplaceWith" )
		lua_getfield( L, Index, "TriggeredOnAction" )
		lua_getfield( L, Index, "ReplaceBlocks" )
		lua_getfield( L, Index, "AdjustToNormal" )
		
		Block.Kind      = lua_tonumber ( L, Index + 1 )
		Block.IsObject  = lua_toboolean( L, Index + 2 )
		Block.IsTrigger = lua_toboolean( L, Index + 3 )
		
		Block.AdjustToNormal = lua_toboolean( L, Index + 9 )
		
		If Block.IsTrigger Then
			Block.TriggeredOnTouch  = lua_toboolean( L, Index + 4 )
			Block.TriggeredBlock    = lua_tonumber ( L, Index + 5 )
			Block.ReplaceWith       = lua_tonumber ( L, Index + 6 )
			Block.TriggeredOnAction = lua_toboolean( L, Index + 7 )
			
			If Not lua_isnil( L, Index + 8 ) Then
				Local Length:Int = lua_objlen( L, Index + 8 )
				
				If Length Mod 2 TServer.AbortServer( "Lua Error: ReplaceBlocks must have even element count" )
				
				Block.ReplaceBlocks = New Int[ Length ]
				
				Local BlocksIndex:Int = lua_gettop( L )
				For Local I:Int = 0 Until Length
					lua_rawgeti( L, Index + 8, I + 1 )
					Block.ReplaceBlocks[ I ] = lua_tonumber( L, BlocksIndex + I + 1 )
				Next
			EndIf
		EndIf
		
		lua_settop( L, Index )
		
		If Not TBlockType.BlockTypes[ Block.Kind ] Then
			TBlockType.BlockTypes[ Block.Kind ] = Block
		Else
			TServer.AbortServer( "Lua Error: Duplicate block ID '" + Block.Kind + "'" )
		EndIf
	End Function
End Type
