SuperStrict

'Import Chaos.Random
Import BRL.Math
Import BRL.Random
Import Pub.ZLib
Import BRL.StandardIO
Import BRL.LinkedList
Import BRL.Bank
Import BRL.FileSystem

Type TGameObject
	Global GameObjects:TList = New TList
	
	Global AdjustToNormalCounter:Int
	
	Field Kind:Int
	
	Field X:Int
	Field Y:Int
	Field Z:Int
	
	Field NX:Float
	Field NY:Float
	Field NZ:Float
	
	Field Link:TLink
	
	Method New()
		Link = GameObjects.AddLast( Self )
	End Method
	
	Function Add( X:Int, Y:Int, Z:Int, NX:Float, NY:Float, NZ:Float, Kind:Int, Map:TVoxelMap )
		Local GO:TGameObject = Map.GameObjects[ X, Y, Z ]
		
		If GO And TBlockType.Get( GO.Kind ).AdjustToNormal Then AdjustToNormalCounter :- 1
		
		If Not GO Then GO = New TGameOBject
		
		GO.X = X
		GO.Y = Y
		GO.Z = Z
		GO.NX = NX
		GO.NY = NY
		GO.NZ = NZ
		GO.Kind = Kind
		
		If TBlockType.Get( Kind ).AdjustToNormal Then AdjustToNormalCounter :+ 1
		
		Map.GameObjects[ X, Y, Z ] = GO
	End Function
	
	Function Remove( X:Int, Y:Int, Z:Int, Map:TVoxelMap )
		If Map.GameObjects[ X, Y, Z ] Then
			If TBlockType.Get( Map.Data[ X, Y, Z ] ).AdjustToNormal Then AdjustToNormalCounter :- 1
			
			Map.GameObjects[ X, Y, Z ].Link.Remove()
			
			Map.GameObjects[ X, Y, Z ] = Null
		EndIf
	End Function
End Type

Type TBlockType
	Const MAX_BLOCK_COUNT:Int = 1024
	
	Global BlockTypes:TBlockType[ MAX_BLOCK_COUNT ]
	
	Field Kind:Int
	
	Field AdjustToNormal:Int
	
	Field IsObject:Int
	Field IsTrigger:Int
	
	Field TriggeredOnTouch:Int
	Field TriggeredOnAction:Int
	Field TriggeredBlock:Int
	Field ReplaceWith:Int
	Field ReplaceBlocks:Int[]
	
	Function Get:TBlockType( I:Int )
		Return BlockTypes[ I ]
	End Function
End Type

Type TScheduledBlock
	Global Blocks:TList = New TList
	
	Field X:Int
	Field Y:Int
	Field Z:Int
	Field BlockType:Int
	Field Map:TVoxelMap
	
	Field Timer:Int
	
	Method New()
		Blocks.AddLast( Self )
	End Method
	
	Function Update()
		For Local Block:TScheduledBlock = EachIn Blocks
			If Block.Timer < MilliSecs() Then
				If Not Block.Map.SaveAccess( Block.X, Block.Y, Block.Z ) Then
					Block.Map.Modify( Block.X, Block.Y, Block.Z, 0.0, 0.0, 0.0, Block.BlockType )
					
					If Block.Map.SaveAccess( Block.X + 1, Block.Y, Block.Z ) = 0 Then Create( Block.X + 1, Block.Y, Block.Z, Block.BlockType, TVoxelMap.WATER_UPDATE_DELAY, Block.Map )
					If Block.Map.SaveAccess( Block.X - 1, Block.Y, Block.Z ) = 0 Then Create( Block.X - 1, Block.Y, Block.Z, Block.BlockType, TVoxelMap.WATER_UPDATE_DELAY, Block.Map )
					If Block.Map.SaveAccess( Block.X, Block.Y - 1, Block.Z ) = 0 Then Create( Block.X, Block.Y - 1, Block.Z, Block.BlockType, TVoxelMap.WATER_UPDATE_DELAY, Block.Map )
					If Block.Map.SaveAccess( Block.X, Block.Y, Block.Z + 1 ) = 0 Then Create( Block.X, Block.Y, Block.Z + 1, Block.BlockType, TVoxelMap.WATER_UPDATE_DELAY, Block.Map )
					If Block.Map.SaveAccess( Block.X, Block.Y, Block.Z - 1 ) = 0 Then Create( Block.X, Block.Y, Block.Z - 1, Block.BlockType, TVoxelMap.WATER_UPDATE_DELAY, Block.Map )
				EndIf
				
				Blocks.Remove( Block )
			EndIf
		Next
	End Function
	
	Function Create:TScheduledBlock( X:Int, Y:Int, Z:Int, BlockType:Int, Delay:Int, Map:TVoxelMap )
		Local Block:TScheduledBlock = New TScheduledBlock
		
		Block.X = X
		Block.Y = Y
		Block.Z = Z
		Block.BlockType = BlockType
		Block.Map = Map
		Block.Timer = MilliSecs() + Delay
		
		Return Block
	End Function
End Type

Type TVoxelMap
	Const BLOCK_GRASS :Byte = 1
	Const BLOCK_DIRT  :Byte = 2
	Const BLOCK_STONE :Byte = 6
	Const BLOCK_WATER :Byte = 9
	Const BLOCK_SNOW  :Byte = 10
	
	Const VEGETATION_START:Byte = 20
	Const VEGETATION_END  :Byte = 23
	
	Const WATER_UPDATE_DELAY:Int = 300
	
	Const MAX_REPLACE_RANGE:Int = 16
	
	Field Width:Int
	Field Height:Int
	Field Depth:Int
	Field GameObjects:TGameObject[,,]
	Field Data:Byte[,,]
	
	Field Modified:Int
	
	Method ClearModifiedFlag()
		Modified = False
	End Method
	
	Function LoadMap:TVoxelMap( File:String )
		Local Instream:TStream = ReadFile( File )
		
		If Instream Then
			Local Size:Int = Instream.ReadInt()
			Local Width:Int = Instream.ReadInt()
			Local Height:Int = Instream.ReadInt()
			Local Depth:Int = Instream.ReadInt()
			
			Local Map:TVoxelMap = Create( Width, Height, Depth )
			
			Map.Load( Size, Instream )
			
			Return Map
		Else
			Return Null
		EndIf
	End Function
	
	Method Load:Int( Size:Int, Instream:TStream )	
		Local Source:Byte Ptr = MemAlloc( Size )
		Instream.ReadBytes( Source, Size )
		
		Local DestSize:Int = Width*Height*Depth
		
		uncompress( Varptr Data[ 0, 0, 0 ], DestSize, Source, Size )
		
		MemFree( Source )
		
		Local NormalGOCount:Int = Instream.ReadInt()
		
		For Local I:Int = 0 Until NormalGOCount
			Local X:Int = Instream.ReadInt()
			Local Y:Int = Instream.ReadInt()
			Local Z:Int = Instream.ReadInt()
			Local NX:Float = Instream.ReadFloat()
			Local NY:Float = Instream.ReadFloat()
			Local NZ:Float = Instream.ReadFloat()
			
			TGameObject.Add( X, Y, Z, NX, NY, NZ, Data[ X, Y, Z ], Self )
		Next
		
		Instream.Close()
		
		For Local X:Int = 0 Until Width
			For Local Y:Int = 0 Until Height
				For Local Z:Int = 0 Until Depth
					If TBlockType.Get( Data[ X, Y, Z ] ).IsObject And Not GameObjects[ X, Y, Z ] Then ..
						TGameObject.Add( X, Y, Z, 0.0, 0.0, 0.0, Data[ X, Y, Z ], Self )
				Next
			Next
		Next
	End Method
	
	Method Save:Int( FileName:String, ZippedData:Byte Ptr, ZippedDataSize:Int )
		Local Outstream:TStream = WriteFile( FileName )
		
		If Outstream Then
			Outstream.WriteInt( ZippedDataSize )
			Outstream.WriteInt( Width )
			Outstream.WriteInt( Height )
			Outstream.WriteInt( Depth )
			Outstream.WriteBytes( ZippedData, ZippedDataSize )
			Outstream.WriteInt( TGameObject.AdjustToNormalCounter )
			
			If TGameObject.AdjustToNormalCounter Then
				For Local GO:TGameObject = EachIn TGameObject.GameObjects
					If TBlockType.Get( GO.Kind ).AdjustToNormal Then
						Outstream.WriteInt( GO.X )
						Outstream.WriteInt( GO.Y )
						Outstream.WriteInt( GO.Z )
						Outstream.WriteFloat( GO.NX )
						Outstream.WriteFloat( GO.NY )
						Outstream.WriteFloat( GO.NZ )
					EndIf
				Next
			EndIf
			
			Outstream.Close()
		Else
			Return False
		EndIf
		
		Return True
	End Method
	
	Method Interact( X:Int, Y:Int, Z:Int )
		Local BlockType:TBlockType = TBlockType.Get( Data[ X, Y, Z ] )
		
		If BlockType.IsTrigger Then
			Local GO:TGameObject = GameObjects[ X, Y, Z ]
			
			If GO Then
				If BlockType.TriggeredBlock Then TriggerEachKind( BlockType.TriggeredBlock )
				If BlockType.ReplaceWith Then Modify( X, Y, Z, GO.NX, GO.NY, GO.NZ, BlockType.ReplaceWith )
				If BlockType.ReplaceBlocks Then ReplaceBlocks( X, Y, Z, BlockType.ReplaceBlocks )
				
				Modified = True
			EndIf
		EndIf
	End Method
	
	Method TriggerEachKind( Kind:Int )
		'TServer.AbortServer( "TriggerEachKind currently not implemented" )
	End Method
	
	Method ReplaceBlocks( CenterX:Int, CenterY:Int, CenterZ:Int, Blocks:Int[] )
		For Local X:Int = Max( 0, CenterX - MAX_REPLACE_RANGE ) Until Min( Width, CenterX + MAX_REPLACE_RANGE )
			For Local Y:Int = Max( 0, CenterY - MAX_REPLACE_RANGE ) Until Min( Height, CenterY + MAX_REPLACE_RANGE )
				For Local Z:Int = Max( 0, CenterZ - MAX_REPLACE_RANGE ) Until Min( Depth, CenterZ + MAX_REPLACE_RANGE )
					For Local I:Int = 0 Until Blocks.Length Step 2
						If Data[ X, Y, Z ] = Blocks[ I ] Then
							Local GO:TGameObject = GameObjects[ X, Y, Z ]
							
							If GO Then
								Modify( X, Y, Z, GO.NX, GO.NY, GO.NZ, Blocks[ I + 1 ] )
							Else
								Modify( X, Y, Z, 0.0, 0.0, 0.0, Blocks[ I + 1 ] )
							EndIf
							
							Exit
						EndIf
					Next
				Next
			Next
		Next
	End Method
	
	Method Modify( X:Int, Y:Int, Z:Int, NX:Float, NY:Float, NZ:Float, NewValue:Byte )
		If TBlockType.Get( Data[ X, Y, Z ] ).IsObject And NewValue = 0 Then TGameObject.Remove( X, Y, Z, Self )
		If TBlockType.Get( NewValue ).IsObject Then TGameObject.Add( X, Y, Z, NX, NY, NZ, NewValue, Self )
		
		If X >= 0 And Y >= 0 And Z >= 0 And X < Width And Y < Height And Z < Depth Then
			Data[ X, Y, Z ] = NewValue
		Else
			Return
		EndIf
		
		Modified = True
		
		If NewValue = 0 Then
			Local GeneratorCount:Int, WallCount:Int
			For Local I:Int = -1 To 1 Step 2
				GeneratorCount :+ ( SaveAccess( X + I, Y, Z ) = BLOCK_WATER )
				GeneratorCount :+ ( SaveAccess( X, Y + 1, Z ) = BLOCK_WATER )
				GeneratorCount :+ ( SaveAccess( X, Y, Z + I ) = BLOCK_WATER )
				
				WallCount :+ ( SaveAccess( X + I, Y, Z ) <> 0 )
				WallCount :+ ( SaveAccess( X, Y - 1, Z ) <> 0 )
				WallCount :+ ( SaveAccess( X, Y, Z + I ) <> 0 )
			Next
			
			If WallCount = 6 And GeneratorCount > 0 Then
				Data[ X, Y, Z ] = BLOCK_WATER
			ElseIf GeneratorCount > 0 Then
				TScheduledBlock.Create( X, Y, Z, BLOCK_WATER, WATER_UPDATE_DELAY, Self )
			EndIf
		EndIf
	End Method
	
	Method SaveAccess:Int( X:Int, Y:Int, Z:Int )
		If X >= 0 And Y >= 0 And Z >= 0 And X < Width And Y < Height And Z < Depth Then Return Data[ X, Y, Z ]
		
		Return -1
	End Method
	
	Method Generate( Seed:Int )
		SeedRnd Seed
		
		Local HeightMap:Float[,] = DiamondSquare( Width, Rnd( 0, 1 ), Rnd( 0, 1 ), Rnd( 0, 1 ), Rnd( 0, 1 ) )
		
		Local MinValue:Float = 1000, MaxValue:Float = -1000
		For Local X:Int = 0 Until Width
			For Local Z:Int = 0 Until Depth
				MinValue = Min( HeightMap[ X, Z ], MinValue )
				MaxValue = Max( HeightMap[ X, Z ], MaxValue )
			Next
		Next
		
		Local Median:Int = Height*0.25
		
		For Local X:Int = 0 Until Width
			For Local Z:Int = 0 Until Depth
				For Local Y:Int = 0 Until Height - 1
					Local MapH:Int = Floor( Sin( 90.0*( HeightMap[ X, Z ] - MinValue )/( MaxValue - MinValue ) )*Height*0.95 )
					
					If Y = MapH Then
						If Y > Median + 10 Then
							Data[ X, Y, Z ] = BLOCK_SNOW
						Else
							Data[ X, Y, Z ] = BLOCK_GRASS
							
							If Y >= Median And Rand( 1, 30 ) = 1 Then
								Data[ X, Y + 1, Z ] = Rand( VEGETATION_START, VEGETATION_END )
								
								TGameObject.Add( X, Y + 1, Z, 0.0, 0.0, 0.0, Data[ X, Y + 1, Z ], Self )
							EndIf
						EndIf
					ElseIf Y < MapH Then
						Data[ X, Y, Z ] = BLOCK_DIRT
					ElseIf Y < Median Then
						Data[ X, Y, Z ] = BLOCK_WATER
					EndIf
				Next
			Next
		Next
	End Method
	
	Function DiamondSquare:Float[,]( Size:Int, Seed1:Float, Seed2:Float, Seed3:Float, Seed4:Float )
		Local HeightMap:Float[ Size + 1, Size + 1 ]
		
		HeightMap[    0,    0 ] = Seed1
		HeightMap[ Size,    0 ] = Seed2
		HeightMap[ Size, Size ] = Seed3
		HeightMap[    0, Size ] = Seed4
		
		Local InvSize:Float = 1.0/Float( Size )
		
		Local StepSize:Int = Size
		While StepSize > 1
			Local HalfStepSize:Int = StepSize Shr 1
			
			For Local X:Int = 0 To Size - StepSize
				For Local Y:Int = 0 To Size - StepSize
					HeightMap[ X + HalfStepSize, Y                ] = ( HeightMap[ X           , Y            ] + HeightMap[ X + StepSize, Y            ] )*0.5
					HeightMap[ X               , Y + HalfStepSize ] = ( HeightMap[ X           , Y            ] + HeightMap[ X           , Y + StepSize ] )*0.5
					HeightMap[ X + HalfStepSize, Y +     StepSize ] = ( HeightMap[ X           , Y + StepSize ] + HeightMap[ X + StepSize, Y + StepSize ] )*0.5
					HeightMap[ X +     StepSize, Y + HalfStepSize ] = ( HeightMap[ X + StepSize, Y            ] + HeightMap[ X + StepSize, Y + StepSize ] )*0.5
					
					HeightMap[ X + HalfStepSize, Y + HalfStepSize ] = ( HeightMap[ X, Y ] + HeightMap[ X + StepSize, Y ] + HeightMap[ X, Y + StepSize ] + HeightMap[ X + StepSize, Y + StepSize ] )*0.25
					HeightMap[ X + HalfStepSize, Y + HalfStepSize ] = HeightMap[ X + HalfStepSize, Y + HalfStepSize ] + Rnd( -1.0, 1.0 )*StepSize*InvSize
					
					Y = Y + StepSize - 1
				Next
				
				X = X + StepSize - 1
			Next
			
			StepSize = HalfStepSize
		Wend
		
		Return HeightMap
	End Function
	
	Function Create:TVoxelMap( Width:Int, Height:Int, Depth:Int )
		Local Map:TVoxelMap = New TVoxelMap
		
		Map.Width = Width
		Map.Height = Height
		Map.Depth = Depth
		Map.Data = New Byte[ Width, Height, Depth ]
		Map.GameObjects = New TGameObject[ Width, Height, Depth ]
		Map.Modified = True
		
		Return Map
	End Function
End Type
