SuperStrict

?Win32
	Framework BRL.GNet
	Import BRL.System
?Not Win32
	Framework BRL.Blitz
?
Import BRL.StandardIO
Import BRL.Socket
Import BRL.SocketStream
Import BRL.Threads
Import BRL.LinkedList
Import BRL.Map
Import Pub.Lua

Import "ENet.bmx"
Import "Configuration.bmx"
Import "Map.bmx"
Include "AccountSystem.bmx"
Include "BanSystem.bmx"
Include "Time.bmx"
Include "Player.bmx"
Include "Lua.bmx"
Include "Logger.bmx"

TServer.Init()

Type TServer
	Const NETMSG_CONNECT        :Int = 1
	Const NETMSG_KEYHIT         :Int = 2
	Const NETMSG_POSITION       :Int = 3
	Const NETMSG_MODIFY_BLOCK   :Int = 4
	Const NETMSG_KICK           :Int = 5
	Const NETMSG_DISCONNECT     :Int = 6
	Const NETMSG_CHAT           :Int = 7
	Const NETMSG_START_GAME     :Int = 8
	Const NETMSG_RESPOND_PING   :Int = 9
	Const NETMSG_PING_LIST      :Int = 10
	Const NETMSG_GENERATE_MAP   :Int = 11
	Const NETMSG_GODMODE        :Int = 12
	Const NETMSG_DONE_GENERATING:Int = 13
	Const NETMSG_MAPDATA        :Int = 14
	Const NETMSG_MAPHEADER      :Int = 15
	Const NETMSG_INTERACTION    :Int = 16
	Const NETMSG_CROUCHING      :Int = 17
	Const NETMSG_DENIED         :Int = 18
	Const NETMSG_LOGIN          :Int = 19
	
	Const NETWORK_POSITION_DELAY:Int = 200
	Const NETWORK_PING_DELAY    :Int = 1000
	
	Const SUPPORTED_VERSION:Int = $000206
	Const VERSION_STRING:String = "0.2.6"
	
	Const MAX_SEND_STEP:Int = 5120
	
	Global PingTimer:Int
	
	Global Config:TConfiguration
	
	Global Host:TENetHost
	Global TCPSocket:TSocket
	
	Global CurrentMap:TVoxelMap
	
	Global LogFile:TStream
	
	Global MapMutex:TMutex, LogMutex:TMutex
	
	Function Init()
		AppendLog( "Initializing ENet..." )
		If ENet_Initialize() Then AbortServer( "Failed to intialize ENet!" )
		
		AppendLog( "Creating TCP socket..." )
		TCPSocket = CreateTCPSocket()
		If Not TCPSocket Then AbortServer( "Failed to create socket!" )
		
		OnEnd( ENet_DeInitialize )
		
		AppendLog( "Loading config..." )
		TConfigurationService.LoadConfigFile()
		Config = TConfigurationService.GetConfiguration()
		
		AppendLog( "Loading lua..." )
		TLuaImporter.Init()
		
		AppendLog( "Initializing logging..." )
		TLogger.Init()
		
		AppendLog( "Loading account data..." )
		TAccountSystem.Init()
		
		AppendLog( "Loading ban list..." )
		TBanSystem.Init()
		
		Local ServerAddress:TENetAddress = TENetAddress.Create( TENetAddress.ENET_HOST_ANY, Config.Port )
		
		AppendLog( "Creating host..." )
		Host = TENetHost.Create( ServerAddress, Config.MaxPlayers, Config.InBandwidth, Config.OutBandwidth )
		
		If Not Host Then AbortServer( "Failed to create host!" )
		
		AppendLog( "Setting up threading environment..." )
		TCPSocket.Bind( Config.Port )
		TCPSocket.Listen( 0 )
		
		MapMutex = CreateMutex()
		LogMutex = CreateMutex()
		CreateThread( CheckSocket, Null )
		CreateThread( AutoSave, Config )
		
		If Config.LoadMap Then
			AppendLog( "Loading map..." )
			
			CurrentMap = TVoxelMap.LoadMap( Config.LoadMap )
			
			If Not CurrentMap Then
				AbortServer( "Map " + Config.LoadMap + " not found." )
			Else
				Config.MapWidth = CurrentMap.Width
				Config.MapHeight = CurrentMap.Height
				Config.MapDepth = CurrentMap.Depth
			EndIf
		Else
			AppendLog( "Generating map..." )
			
			CurrentMap = TVoxelMap.Create( Config.MapWidth, Config.MapHeight, Config.MapDepth )
			CurrentMap.Generate( MilliSecs() )
		EndIf
		
		AppendLog( "Server fired up and ready to go" )
		
		PingTimer = MilliSecs()
		
		While True
			UpdateENet()
			
			HandleInput()
			
			TScheduledBlock.Update()
		Wend
	End Function
	
	Function AutoSave:Object( Data:Object )
		Local Timeout:Int = TConfiguration( Data ).AutoSaveDelay
		
		If Not Timeout Then Return Null
		
		Repeat
			Delay Timeout*1000
			
			SaveMapIfModified()
		Forever
	End Function
	
	Function ThreadedSave:Object( Data:Object )
		If Not SaveMapIfModified() Then AppendLog( "Save ignored. Map has not been modified since last save." )
		SaveLogs()
	End Function
	
	Function SaveMapIfModified:Int()
		If CurrentMap.Modified Then
			Local FileName:String = "Maps/" + GetTimeString().Replace( ":", "-" ) + ".dat"
			
			Local CompressionTime:Int = MilliSecs()
			MapMutex.Lock()
				Local MapSize:Int
				Local Data:Byte Ptr = ZipMap( MapSize )
			MapMutex.UnLock()
			CompressionTime = MilliSecs() - CompressionTime
			
			Local Result:Int = CurrentMap.Save( FileName, Data, MapSize )
			
			MemFree( Data )
			
			Config.LoadMap = FileName
			TConfigurationService.SaveConfigFile()
			
			If Result Then
				AppendLog( "Saved map to /" + FileName + ". Filesize: " + FileSize( FileName )/1024 + "kb. Compression time: " + CompressionTime + "ms." )
			Else
				AppendLog( "Failed to save map " + FileName )
			EndIf
			
			CurrentMap.ClearModifiedFlag()
			
			Return True
		Else
			Return False
		EndIf
	End Function
	
	Function SaveLogs()
		Local Filename:String = GetTimeString().Replace( ":", "-" ) + ".log"
		
		If Not TLogger.Save( "Logs/" + Filename ) Then
			AppendLog( "Failed to save log " + Filename )
		Else
			AppendLog( "Saved log to " + FileName + ". Filesize: " + FileSize( FileName )/1024 + "kb." )
			
			TLogger.Clear()
		EndIf
	End Function
	
	Function CheckSocket:Object( Data:Object )
		Repeat
			Local RemoteSocket:TSocket = TCPSocket.Accept( 50 )
			
			If RemoteSocket Then CreateThread( SendMapThreaded, RemoteSocket )
		Forever
	End Function
	
	Function ZipMap:Byte Ptr( MapSize:Int Var )
		Local Size:Int = CurrentMap.Width*CurrentMap.Height*CurrentMap.Depth
		
		MapSize = Size
		
		Local Data:Byte Ptr = MemAlloc( Size )
		Local CompressedData:Byte Ptr = MemAlloc( Size )
		
		MapMutex.Lock()
			MemCopy( Data, Varptr CurrentMap.Data[ 0, 0, 0 ], Size )
		MapMutex.UnLock()
		
		compress2( CompressedData, MapSize, Data, Size, 7 )
		
		MemFree Data
		
		Return CompressedData
	End Function
	
	Function SendMapThreaded:Object( Data:Object )
		Local RemoteSocket:TSocket = TSocket( Data )
		
		Try
			Local CompressionTime:Int = MilliSecs()
			
			Local MapSize:Int
			Local Data:Byte Ptr = ZipMap( MapSize )
			
			CompressionTime = MilliSecs() - CompressionTime
			
			Local Outstream:TStream = CreateSocketStream( RemoteSocket, True )
			Outstream.WriteInt( MapSize )
			
			For Local I:Int = 0 Until MapSize Step MAX_SEND_STEP
				Outstream.WriteBytes( Data + I, Min( MapSize - I, MAX_SEND_STEP ) )
				Outstream.Flush()
				
				Delay 100
			Next
			
			MemFree( Data )
			
			Outstream.WriteInt( TGameObject.AdjustToNormalCounter )
			If TGameObject.AdjustToNormalCounter Then
				Local MemorySize:Int = TGameObject.AdjustToNormalCounter*24
				Local OriginalMemory:Byte Ptr = MemAlloc( MemorySize )
				Local NormalBank:Int Ptr = Int Ptr OriginalMemory
				
				For Local GO:TGameObject = EachIn TGameObject.GameObjects
					If TBlockType.Get( GO.Kind ).AdjustToNormal Then
						NormalBank[ 0 ] = GO.X
						NormalBank[ 1 ] = GO.Y
						NormalBank[ 2 ] = GO.Z
						NormalBank[ 3 ] = FloatToInt( GO.NX )
						NormalBank[ 4 ] = FloatToInt( GO.NY )
						NormalBank[ 5 ] = FloatToInt( GO.NZ )
						
						NormalBank :+ 6
					EndIf
				Next
				
				For Local I:Int = 0 Until MemorySize Step MAX_SEND_STEP
					Outstream.WriteBytes( OriginalMemory + I, Min( MemorySize - I, MAX_SEND_STEP ) )
					Outstream.Flush()
					
					Delay 100
				Next
				
				MemFree( OriginalMemory )
			EndIf
			
			AppendLog( "Compression time: " + CompressionTime + "ms. Compressed map size: " + MapSize/1024 + "kb." )
		Catch Exception:Object
			AppendLog( "Lost TCP connection to player." )
		EndTry
	End Function
	
	Function UpdateENet()
		While Host.Service( 50 ) > 0
			Local Event:TENetEvent = Host.GetLastEvent()
			
			Select Event.EventType
				Case Event.ENET_EVENT_TYPE_RECEIVE
					Local  Inpacket:TENetPacket = Event.Packet
					Local Outpacket:TENetPacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
					
					Local  Instream:TENetStream = TENetStream.Create(  Inpacket )
					Local Outstream:TENetStream = TENetStream.Create( Outpacket )
					
					Local MessageType:Int = Instream.ReadByte()
					
					Select MessageType
						Case NETMSG_CONNECT
							DispatchConnect( Inpacket, Outpacket, Instream, Outstream, Event )
						Case NETMSG_KEYHIT
							DispatchKeyhit( Inpacket, Outpacket, Instream, Outstream )
						Case NETMSG_POSITION
							DispatchPosition( Inpacket, Outpacket, Instream, Outstream )
						Case NETMSG_MODIFY_BLOCK
							DispatchModifyBlock( Inpacket, Outpacket, Instream, Outstream, Event )
						Case NETMSG_CHAT
							DispatchChat( Inpacket, Outpacket, Instream, Outstream, Event )
						Case NETMSG_GODMODE
							DispatchGodMode( Inpacket, Outpacket, Instream, Outstream )
						Case NETMSG_CROUCHING
							DispatchCrouching( Inpacket, Outpacket, Instream, Outstream )
						Case NETMSG_DONE_GENERATING
							DispatchDoneGenerating( Inpacket, Outpacket, Instream, Outstream, Event )
						Case NETMSG_INTERACTION
							DispatchInteraction( Inpacket, Outpacket, Instream, Outstream, Event )
					End Select
				Case Event.ENET_EVENT_TYPE_DISCONNECT
					DispatchDisconnect( Event )
			End Select
		Wend
		
		If MilliSecs() - PingTimer > NETWORK_PING_DELAY Then
			SendPingList()
			
			PingTimer = MilliSecs()
		EndIf
	End Function
	
	Function DispatchConnect( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream, Event:TENetEvent )
		Local Length:Int = Instream.ReadInt()
		
		Local Name:String = Instream.ReadString( Length )
		
		If TPlayer.FindPlayerByName( Name ) Then
			DenyPlayer( "Nickname is already in use.", Event, Outstream, Outpacket )
			
			Return
		ElseIf Name.ToLower() = "horst" Or Name.Find( " " ) >= 0 Or Name = "" Then
			DenyPlayer( "Forbidden nickname. Please rename your player.", Event, Outstream, Outpacket )
			
			Return
		EndIf
		
		Local Version:Int = Instream.ReadInt()
		
		If Version <> SUPPORTED_VERSION Then
			DenyPlayer( "Incompatible client version. Supported version is " + VERSION_STRING + ".", Event, Outstream, Outpacket )
			
			Return
		EndIf
		
		Local Player:TPlayer = TPlayer.Create( Event.Peer )
		
		If TBanSystem.PlayerBanned( Player ) Then
			DenyPlayer( "You have been banned from this server.", Event, Outstream, Outpacket )
			
			Player.Remove()
			
			Return
		EndIf
		
		Player.SetName( Name )
		
		AppendLog( "New Player: " + Player.Name + " Host: " + Player.Peer.Address.ToString() + " Port: " + Player.Peer.Address.Port )
		
		TLogger.AddPlayer( Player )
		
		For Local CurrPlayer:TPlayer = EachIn TPlayer.Players
			If CurrPlayer = TPlayer.Players.Last() Then Exit
			
			Outpacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
			Outstream = TENetStream.Create( Outpacket )
			
			Outstream.WriteByte( NETMSG_CONNECT )
			Outstream.WriteInt( 0 )
			Outstream.WriteInt( CurrPlayer.ID )
			Outstream.WriteByte( 1 )
			Outstream.WriteInt( Player.Name.Length )
			Outstream.WriteString( Player.Name )
			Outstream.WriteInt( Player.ID )
			Outstream.WriteInt( Player.GodMode )
			Outstream.WriteInt( Player.Crouching )
			
			CurrPlayer.Peer.Send( Outpacket )
		Next
		
		Outpacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
		Outstream = TENetStream.Create( Outpacket )
		Outstream.WriteByte( NETMSG_CONNECT )
		Outstream.WriteInt( 0 )
		Outstream.WriteInt( Player.ID )
		Outstream.WriteByte( TPlayer.Players.Count() - 1 )
		
		For Local CurrPlayer:TPlayer = EachIn TPlayer.Players
			If CurrPlayer = TPlayer.Players.Last() Then Exit
			
			Outstream.WriteInt( CurrPlayer.Name.Length )
			Outstream.WriteString( CurrPlayer.Name )
			Outstream.WriteInt( CurrPlayer.ID )
			Outstream.WriteInt( CurrPlayer.GodMode )
			Outstream.WriteInt( CurrPlayer.Crouching )
		Next
		
		Player.Peer.Send( Outpacket )
		
		Outpacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
		Outstream = TENetStream.Create( Outpacket )
		Outstream.WriteByte( NETMSG_MAPDATA )
		Outstream.WriteInt( Config.MapWidth )
		Outstream.WriteInt( Config.MapHeight )
		Outstream.WriteInt( Config.MapDepth )
		Outstream.WriteInt( MilliSecs() )
		
		Player.Peer.Send( Outpacket )
	End Function
	
	Function DispatchKeyhit( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream )
		Local PeerID:Int = Instream.ReadInt()
		Local Input:Int  = Instream.ReadInt()
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PeerID )
		
		Player.SetInput( Input )
		
		Outstream.WriteByte( NETMSG_KEYHIT )
		Outstream.WriteInt( PeerID )
		Outstream.WriteInt( Input )
		
		SendToEachPlayer( Outpacket, Player )
	End Function
	
	Function DispatchPosition( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream )
		Local PeerID:Int = Instream.ReadInt()
		Local PlayerX:Float = Instream.ReadFloat()
		Local PlayerY:Float = Instream.ReadFloat()
		Local PlayerZ:Float = Instream.ReadFloat()
		Local RotX:Float    = Instream.ReadFloat()
		Local RotY:Float    = Instream.ReadFloat()
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PeerID )
		
		Player.SetCoordinates( PlayerX, PlayerY, PlayerZ, RotX, RotY )
		
		Outstream.WriteByte( NETMSG_POSITION )
		Outstream.WriteInt( PeerID )
		Outstream.WriteFloat( PlayerX )
		Outstream.WriteFloat( PlayerY )
		Outstream.WriteFloat( PlayerZ )
		Outstream.WriteFloat( RotX )
		Outstream.WriteFloat( RotY )
		
		SendToEachPlayer( Outpacket, Player )
	End Function
	
	Function DispatchModifyBlock( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream, Event:TENetEvent )
		Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
		
		If Not Player.IsRegistered Then Return
		
		Local VoxelX  :Int = Instream.ReadInt()
		Local VoxelY  :Int = Instream.ReadInt()
		Local VoxelZ  :Int = Instream.ReadInt()
		Local NormalX :Float = Instream.ReadFloat()
		Local NormalY :Float = Instream.ReadFloat()
		Local NormalZ :Float = Instream.ReadFloat()
		Local NewBlock:Int = Instream.ReadInt()
		
		Outstream.WriteByte( NETMSG_MODIFY_BLOCK )
		Outstream.WriteInt( VoxelX )
		Outstream.WriteInt( VoxelY )
		Outstream.WriteInt( VoxelZ )
		Outstream.WriteFloat( NormalX )
		Outstream.WriteFloat( NormalY )
		Outstream.WriteFloat( NormalZ )
		Outstream.WriteInt( NewBlock )
		
		SendToEachPlayer( Outpacket, Player )
		
		MapMutex.Lock()
			CurrentMap.Modify( VoxelX, VoxelY, VoxelZ, NormalX, NormalY, NormalZ, NewBlock )
		MapMutex.UnLock()
		
		TLogger.LogMapModify( Player, VoxelX, VoxelY, VoxelZ, NewBlock )
	End Function
	
	Function DispatchDisconnect( Event:TENetEvent )
		Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
		
		If Not Player Then Return
		
		AppendLog( "Player disconnected: " + Player.Name )
		
		TLogger.RemovePlayer( Player )
		
		Player.Remove()
		
		Local Outpacket:TENetPacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
		Local Outstream:TENetStream = TENetStream.Create( Outpacket )
		
		Outstream.WriteByte( NETMSG_DISCONNECT )
		Outstream.WriteInt( Player.ID )
		
		SendToEachPlayer( Outpacket )
	End Function
	
	Function DispatchChat( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream, Event:TENetEvent )
		Local Length:Int = Instream.ReadInt()
		Local Message:String = Instream.ReadString( Length )
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
		
		If Message[ 0 ] = "/"[ 0 ] Then
			Local OutMessage:String = ParseCommand( Message, Event, Outstream, Outpacket )
			
			If OutMessage Then
				Outpacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
				Outstream = TENetStream.Create( Outpacket )
				
				Outstream.WriteByte( NETMSG_CHAT )
				Outstream.WriteInt( Outmessage.Length )
				Outstream.WriteString( Outmessage )
				
				Player.Peer.Send( Outpacket )
				
				AppendLog( "(to " + Player.Name + "): " + Outmessage )
			EndIf
		Else
			Outstream.WriteByte( NETMSG_CHAT )
			Outstream.WriteInt( Message.Length )
			Outstream.WriteString( Message )
			Outstream.Seek( Outstream.Pos() - 1 )
			
			Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
			
			SendToEachPlayer( Outpacket, Player )
			
			AppendLog( Message )
		EndIf
	End Function
	
	Function DispatchGodMode( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream )
		Local PeerID:Int = Instream.ReadInt()
		Local Mode  :Int = Instream.ReadInt()
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PeerID )
		Player.SetGodMode( Mode )
		
		Outstream.WriteByte( NETMSG_GODMODE )
		Outstream.WriteInt( PeerID )
		Outstream.WriteInt( Mode )
		
		SendToEachPlayer( Outpacket, Player )
	End Function
	
	Function DispatchDoneGenerating( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream, Event:TENetEvent )
		Outstream.WriteByte( NETMSG_START_GAME )
		
		Event.Peer.Send( OutPacket )
	End Function
	
	Function DispatchInteraction( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream, Event:TENetEvent )
		Local VoxelX:Int = Instream.ReadInt()
		Local VoxelY:Int = Instream.ReadInt()
		Local VoxelZ:Int = Instream.ReadInt()
		
		Outstream.WriteByte( NETMSG_INTERACTION )
		Outstream.WriteInt( VoxelX )
		Outstream.WriteInt( VoxelY )
		Outstream.WriteInt( VoxelZ )
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
		
		SendToEachPlayer( Outpacket, Player )
		
		MapMutex.Lock()
			CurrentMap.Interact( VoxelX, VoxelY, VoxelZ )
		MapMutex.UnLock()
	End Function
	
	Function DispatchCrouching( Inpacket:TENetPacket, Outpacket:TENetPacket, Instream:TENetStream, Outstream:TENetStream )
		Local PeerID:Int = Instream.ReadInt()
		Local Mode  :Int = Instream.ReadInt()
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PeerID )
		Player.SetCrouching( Mode )
		
		Outstream.WriteByte( NETMSG_CROUCHING )
		Outstream.WriteInt( PeerID )
		Outstream.WriteInt( Mode )
		
		SendToEachPlayer( Outpacket, Player )
	End Function
	
	Function ParseCommand:String( Message:String, Event:TENetEvent, Outstream:TENetStream, OutPacket:TENetPacket )
		Message = Message[ 1 .. ]
		
		Local Parts:String[] = Message.Split( " " )
		
		Local Player:TPlayer = TPlayer.FindPlayerByID( PointerToInt( Event.Peer.Original ) )
		
		Local OutMessage:String = "Incorrect use of '" + Parts[ 0 ] + "'."
		Select Parts[ 0 ].ToLower()
			Case "kick"
				If Player.IsAdmin Then
					If Parts.Length >= 2 Then
						Local PlayerName:String = Message[ Parts[ 0 ].Length .. ]
						
						Local PlayerToKick:TPlayer = TPlayer.FindPlayerByName( PlayerName )
						
						If PlayerToKick And Player <> PlayerToKick Then
							KickPlayer( PlayerToKick, Outstream, OutPacket )
						Else
							OutMessage = "Player " + PlayerName + " not found."
						EndIf
					EndIf
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "savemap"
				If Player.IsAdmin Then
					CreateThread( ThreadedSave, Null )
					
					OutMessage = "Map is being saved. See server log for further details."
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "registerplayer"
				If Player.IsAdmin Then
					If Parts.Length = 3 Then
						Local PlayerName:String = Parts[ 1 ]
						Local Password  :String = Parts[ 2 ]
						
						TAccountSystem.AddPlayer( PlayerName.ToLower(), Password )
						
						OutMessage = "Successfully registered player " + PlayerName + "."
					EndIf
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "unregisterplayer"
				If Player.IsAdmin Then
					If Parts.Length = 2 Then
						Local PlayerName:String = Parts[ 1 ]
						
						TAccountSystem.RemovePlayer( PlayerName.ToLower() )
						
						OutMessage = "Successfully removed player " + PlayerName + "."
					EndIf
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "banplayer"
				If Player.IsAdmin Then
					If Parts.Length = 2 Then
						Local PlayerName:String = Parts[ 1 ]
						
						Local Player:TPlayer = TPlayer.FindPlayerByName( PlayerName )
						
						If Player Then
							TBanSystem.AddBan( Player )
							
							OutMessage = "Successfully banned player " + PlayerName + " with host mask " + ..
							             Player.Peer.Address.ToString() + "."
						Else
							OutMessage = "Player " + PlayerName + " not found."
						EndIf
					EndIf
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "banip"
				If Player.IsAdmin Then
					If Parts.Length = 2 Then
						TBanSystem.AddBanHost( Parts[ 1 ] )
						
						OutMessage = "Successfully banned player with host mask " + Parts[ 1 ] + "."
					EndIf
				Else
					OutMessage = "You do not have sufficient rights to do that."
				EndIf
			Case "adminlogin"
				If Parts.Length = 2 And Parts[ 1 ] = Config.AdminPassword Then
					Player.IsAdmin = True
					
					OutMessage = "Password accepted. You are now in admin mode."
				Else
					OutMessage = "Incorrect password."
				EndIf
			Case "login"
				If Parts.Length = 2 Then
					If TAccountSystem.PlayerExists( Player.Name.ToLower() ) Then
						If TAccountSystem.CorrectPassword( Player.Name.ToLower(), Parts[ 1 ] ) Then
							OutMessage = "You have been successfully logged in."
							
							LoginPlayer( Player, Outstream, Outpacket )
							
							Player.IsRegistered = True
						Else
							OutMessage = "Incorrect password."
						EndIf
					Else
						OutMessage = "You are not registered."
					EndIf
				EndIf
			Case "changepassword"
				If Parts.Length = 2 Then
					If Player.IsRegistered And TAccountSystem.PlayerExists( Player.Name.ToLower() ) Then
						TAccountSystem.ChangePlayerPassword( Player.Name.ToLower(), Parts[ 1 ] )
						
						OutMessage = "Your password has been changed successfully."
					EndIf
				EndIf
			Default
				OutMessage = "Unknown command '" + Parts[ 0 ] + "'."
		End Select
		
		Return OutMessage:String
	End Function
	
	Function DenyPlayer( OutString:String, Event:TENetEvent, Outstream:TENetStream, OutPacket:TENetPacket )
		Outstream.WriteByte( NETMSG_DENIED )
		Outstream.WriteInt( OutString.Length )
		Outstream.WriteString( OutString )
		
		Event.Peer.Send( Outpacket )
	End Function
	
	Function KickPlayer( Player:TPlayer, Outstream:TENetStream, OutPacket:TENetPacket )
		Outstream.WriteByte( NETMSG_KICK )
		Outstream.WriteInt( Player.ID )
		
		SendToEachPlayer( Outpacket )
	End Function
	
	Function LoginPlayer( Player:TPlayer, Outstream:TENetStream, OutPacket:TENetPacket )
		Outstream.WriteByte( NETMSG_LOGIN )
		
		Player.Peer.Send( Outpacket )
	End Function
	
	Function SendPingList()
		Local Outpacket:TENetPacket = TENetPacket.Create( Null, 0, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
		Local Outstream:TENetStream = TENetStream.Create( Outpacket )
		
		Outstream.WriteByte( NETMSG_PING_LIST )
		For Local Player:TPlayer = EachIn TPlayer.Players
			Player.Peer.FromENetPeer()
			
			Outstream.WriteInt( Player.ID )
			Outstream.WriteInt( Player.Peer.Ping )
			Outstream.WriteInt( Player.Peer.InBandwidth )
			Outstream.WriteInt( Player.Peer.OutBandwidth )
			Outstream.WriteFloat( Player.Peer.PacketLoss )
		Next
		
		SendToEachPlayer( Outpacket )
	End Function
	
	Function HandleInput()
		
	End Function
	
	Function SendToEachPlayer( Packet:TENetPacket, Exclude:TPlayer = Null )
		For Local Player:TPlayer = EachIn TPlayer.Players
			If Player <> Exclude Then Player.Peer.Send( Packet )
			
			If Player <> TPlayer.Players.Last() Then ..
				Packet = TENetPacket.Create( Packet.Data, Packet.Size, TENetPacket.ENET_PACKET_FLAG_RELIABLE )
		Next
	End Function
	
	Function AbortServer( Message:String )
		AppendLog( "Server error: " + Message )
		
		Print "Aborting server. Press Enter to continue"
		
		Input( "" )
		
		End
	End Function
	
	Function AppendLog( Message:String )
		If Not LogFile Then LogFile = WriteFile( "Log.txt" )
		
		If LogMutex Then
			LogMutex.Lock()
				Message = "[" + GetTime() + "] " + Message
				
				Print Message
				
				LogFile.WriteLine( Message )
				LogFile.Flush()
			LogMutex.UnLock()
		Else
			Message = "[" + GetTime() + "] " + Message
			
			Print Message
			
			LogFile.WriteLine( Message )
			LogFile.Flush()
		EndIf
	End Function
	
	Function FloatToInt:Int( F:Float )
		Return ( Int Ptr ( Varptr F ) )[ 0 ]
	End Function
	
	Function PointerToInt:Int( Pointer:Byte Ptr )
		Return ( Int Ptr ( Varptr Pointer ) )[ 0 ]
	End Function
End Type